#include <windows.h>
#include <malloc.h>
#include <process.h>
#include <mmsystem.h>
#include <dsound.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <conio.h>

#include <stdio.h>
#include <vbf/vbf_std.h>
#include <mpl/mx_std.h>
#include <mpl/ml_xm.h>
#include <mpl/sys/mem.h>

#define PLAYFREQ 44100

mpl__mixer_t *mixah;

LPDIRECTSOUND8 ds8 = 0;
WAVEFORMATEX format;
DSBUFFERDESC desc;
LPDIRECTSOUNDBUFFER8 buf = 0;
int bufpos;

void *tmpbuf;
int tmpbufsize;
int kill=0;
int playing=0;

double sine_pos = 0;

void
fillwf(WAVEFORMATEX *wf, int ch, int freq, int bits)
{
	wf->wFormatTag      = WAVE_FORMAT_PCM;
	wf->nChannels       = ch;
	wf->nSamplesPerSec  = freq;
	wf->wBitsPerSample  = bits;
	wf->nBlockAlign     = bits * ch >> 3;
	wf->nAvgBytesPerSec = wf->nSamplesPerSec * wf->nBlockAlign;
}

int
init()
{
	if (FAILED(DirectSoundCreate8(0, &ds8, 0)))
	{
		return -1;
	}

	if (FAILED(ds8->lpVtbl->SetCooperativeLevel(ds8, GetForegroundWindow(), DSSCL_NORMAL)))
	{
		return -1;
	}

	kill = 0;

	return 0;
}

void
shutitdown()
{
	if (playing)
	{
		kill = 1;

		while(kill) Sleep(5);
	}

	if (buf)
	{
		IUnknown_Release(buf);
		buf = 0;
	}

	if (ds8)
	{
		IUnknown_Release(ds8);
		ds8 = 0;
	}
}


int
setprimarybufferformat(int ch, int freq, int bits)
{
	LPDIRECTSOUNDBUFFER primbuf;
	WAVEFORMATEX format;
	DSBUFFERDESC desc;


	desc.dwBufferBytes = 0;
	desc.lpwfxFormat   = 0;
	desc.dwFlags       = DSBCAPS_PRIMARYBUFFER;
	desc.dwSize        = sizeof(DSBUFFERDESC);
	desc.dwReserved    = 0;

	if (FAILED(ds8->lpVtbl->CreateSoundBuffer(ds8, &desc, &primbuf, 0)))
	{
		return -1;
	}

	fillwf(&format, ch, freq, bits);

	if (FAILED(primbuf->lpVtbl->SetFormat(primbuf, &format)))
	{
		IUnknown_Release(primbuf);

		return -1;
	}

	IUnknown_Release(primbuf);

	return 0;
}


int
calc_space_left()
{
	int readpos, writepos;

	buf->lpVtbl->GetCurrentPosition(buf, &readpos, &writepos);

	if (bufpos < readpos)
	{
		return readpos - bufpos;
	}

	return desc.dwBufferBytes - bufpos + readpos;
}

int
write_data(void *data, mpl__mem_size_t size)
{
	int result, length1, length2;
	void *address1, *address2;

	result = buf->lpVtbl->Lock(buf, bufpos, size,
							   &address1, &length1,
							   &address2, &length2,
							   0);

	if (result == DSERR_BUFFERLOST)
	{
		buf->lpVtbl->Restore(buf);

		result = buf->lpVtbl->Lock(buf, bufpos, size,
								   &address1, &length1,
								   &address2, &length2,
								   0);
	}

	if (result != DS_OK)
	{
		return 0;
	}

	mpl__mem_copy(data, address1, length1);

	if (length2)
	{
		mpl__mem_copy(((i8*)data) + length1, address2, length2);
	}

	buf->lpVtbl->Unlock(buf, address1, length1, address2, length2);

	return 1;
}

int
fillbuf()
{
	mpl__mixer_mix(mixah, tmpbuf, tmpbufsize);
	
	return 0;
}

void __cdecl
dsthread(void *param)
{
	fillbuf();

	while(!kill)
	{
		if (calc_space_left() > tmpbufsize * format.nBlockAlign)
		{
			write_data(tmpbuf, tmpbufsize * format.nBlockAlign);

			bufpos += tmpbufsize * format.nBlockAlign;
			bufpos %= desc.dwBufferBytes;

			fillbuf();
		}

		Sleep(5);
	}

	kill = 0;
}

int
setparams(int ch, int freq, int bits)
{
	if (buf)
	{
		IUnknown_Release(buf);
		buf = 0;
	}

	fillwf(&format, ch, freq, bits);

	desc.dwBufferBytes = format.nBlockAlign * (int)(freq * 0.1);
	desc.lpwfxFormat   = &format;
	desc.dwFlags       = DSBCAPS_GLOBALFOCUS | DSBCAPS_GETCURRENTPOSITION2;
	desc.dwSize        = sizeof(DSBUFFERDESC);
	desc.dwReserved    = 0;

	if (FAILED(ds8->lpVtbl->CreateSoundBuffer(ds8, &desc, (struct IDirectSoundBuffer**)&buf, 0)))
	{
		return -1;
	}

	tmpbufsize = (desc.dwBufferBytes / format.nBlockAlign) >> 2;

	if (mpl__mem_alloc(format.nBlockAlign * tmpbufsize, &tmpbuf) <= MPL__ERR_GENERIC)
	{
		IUnknown_Release(buf);
		buf = 0;
		return -1;
	}

	SetThreadPriority((HANDLE)_beginthread(dsthread, 0, 0), THREAD_PRIORITY_TIME_CRITICAL);

	return 0;
}

int
play()
{
	if (buf->lpVtbl->Play(buf, 0, 0, DSBPLAY_LOOPING) != DS_OK)
	{
		return -1;
	}

	playing = 1;

	return 0;
}

int
clear()
{
	int result, length;
	char *bufpnt;

	result = buf->lpVtbl->Lock(buf, 0, desc.dwBufferBytes,
							   &bufpnt, &length,
							   0, 0,
							   0);

	if (result == DSERR_BUFFERLOST)
	{
		buf->lpVtbl->Restore(buf);

		result = buf->lpVtbl->Lock(buf, 0, desc.dwBufferBytes,
								   &bufpnt, &length,
								   0, 0,
								   0);
	}

	if (result != DS_OK)
	{
		return 0;
	}

	while(length--)
	{
		bufpnt[length] = 0;
	}

	buf->lpVtbl->Unlock(buf, bufpnt, desc.dwBufferBytes, 0, 0);

	return 1;
}

void
stop()
{
	buf->lpVtbl->Stop(buf);

	playing = 0;
}

int
cmd_cm(void *internal, mpl__snd_mixer_t **mixer, mpl__snd_mixer_opt_t opt)
{
	mpl__mixer_opt_t mopt;
	mpl__mixer_t *mx;
	mpl__snd_dev_output_t out;

	mpl__mx_std_create(&mx);

	mixah = mx;

	mopt.ch_num = opt.ch_num;
	mopt.mem_usage = 0;
	mopt.qual = MPL__MX_HQ;
	mopt.smp_num = opt.smp_num;

	mpl__mixer_set_options(mx, &mopt);
	mpl__mixer_set_vol(mx, 1.0f, 0.5f, 0.5f);

	out.channels = 2;
	out.format = MPL__SND_DEV_FMT_16BIT;
	out.freq = PLAYFREQ;
	out.latency = 0;

	mpl__mixer_set_output_options(mx, &out, 100);

	mpl__mixer_get_interface(mx, mixer);

	return 0;
}

int
cmd_dm(void *internal, mpl__snd_mixer_t *mixer)
{
	return 0;
}

int
cmd_goo(void *internal, mpl__snd_dev_output_t *opt)
{
	opt->channels = 2;
	opt->format = 2;
	opt->freq = PLAYFREQ;
	opt->latency = 0;

	return 0;
}

i16 buffer[2 * 10 * PLAYFREQ];

int
main(int argc, char **argv)
{

	mpl__xm_t xm;
	mpl__mp_t mp;
	vbf_t file;
	mpl__snd_dev_t dev;
	FILE *rg;

	mpl__mem_init();

	if (argc != 2)
	{
		printf("syntax: %s filename.xm\n", argv[0]);
		return 0;
	}

	dev.create_mixer = cmd_cm;
	dev.destroy_mixer = cmd_dm;
	dev.get_output_options = cmd_goo;

	rg = fopen(argv[1], "rb");
	if (!rg)
	{
		printf("file not found\n");
		return 0;
	}
	
	vbf_std_open(&file, rg);
	mpl__xm_load(&file, &xm);
	vbf_close(&file);
	fclose(rg);

	if (mpl__mp_xm_construct(&xm, &mp, 0) < MPL__ERR_OK)
	{
		printf("error\n");
		return 0;
	}

	if (mpl__mp_set_dev(&mp, &dev) < MPL__ERR_OK)
	{
		printf("couldn't set device\n");
		return 0;
	}
	
	if (mpl__mp_set_vol(&mp, 0.6f) < MPL__ERR_OK)
	{
		printf("error\n");
		return 0;
	}

	if (mpl__mp_play(&mp) < MPL__ERR_OK)
	{
		printf("error\n");
		return 0;
	}

	mpl__mp_set_loop(&mp, 4);

	init();
	setparams(2, PLAYFREQ, 16);
	clear();
	play();

	printf("playin'.....\n");

	while(!kbhit()) Sleep(100);
	stop();

	return 0;
}
